;;;-*-Mode: LISP; Package: CCL -*-

; Copyright 1995 Digitool, Inc.


; Turns initialize-instance into instance-initialize so that users can
; call-next-method with args and get the args to :before & :after methods.

(defclass instance-initialize-mixin () ())

(defun stream-cant-do-that (s what)
  (error "Stream ~S is not capable of ~a. "s what))

(defun ill-in (s &rest ignore)
  (declare (ignore ignore))
  (stream-cant-do-that s "input"))

(defun ill-out (s &rest ignore)
  (declare (ignore ignore))
  (stream-cant-do-that s "output"))

(defun ill-bin (s &rest ignore)
  (declare (ignore ignore))
  (stream-cant-do-that s "binary input"))

(defun ill-bout (s &rest ignore)
  (declare (ignore ignore))
  (stream-cant-do-that s "binary output"))

(defun ill-chin (s &rest ignore)
  (declare (ignore ignore))
  (stream-cant-do-that s "character input"))

(defun ill-chout (s &rest ignore)
  (declare (ignore ignore))
  (stream-cant-do-that s "character output"))

(defun stream-is-closed (s &rest ignore)
  (declare (ignore ignore))
  (error "Stream ~S is closed." s))

(defun do-nothing (&rest ignore)
  (declare (ignore ignore)))

(defun closed-stream-op (stream &rest ignore)
  (declare (ignore ignore))
  (error "Stream ~S is not open." stream))

(let*()
(defstruct (sstream (:predicate sstreamp)
                    (:constructor %make-sstream))
  ;;
  (chin #'ill-chin)
  (chout #'ill-chout)
  (bin #'ill-bin)
  (bout #'ill-bout)
  (misc #'do-nothing)
  (string-out #'ill-chout)
  (dirty nil))
)

(defun %close-sstream (s)
  (let* ((closed #'closed-stream-op))
    (setf (sstream-chin s) closed
	  (sstream-chout s) closed
	  (sstream-bin s) closed
	  (sstream-bout s) closed
	  (sstream-string-out s) closed
	  (sstream-misc s) closed))
  nil)

(let*()
(defstruct (concatenated-stream
	     (:include sstream
		    (chin #'concatenated-chin)
		    (bin #'concatenated-bin)
		    (misc #'concatenated-misc))
	     (:constructor
	     make-concatenated-stream (&rest streams &aux (current streams))))
  current
  (streams nil :type list))
)

(macrolet ((in-fun (name fun)
	     `(defun ,name (stream eof-errorp eof-value)
		(do* ((current (concatenated-stream-current stream) (cdr current)))
		     ((null current)
		      (if eof-errorp
			  (error 'end-of-file :stream stream)
			  eof-value))
		  (let* ((stream (car current))
			 (result (,fun stream nil nil)))
		    (when result (return result)))
		  (setf (concatenated-stream-current stream) current)))))
  (in-fun concatenated-chin read-char)
  (in-fun concatenated-bin read-byte))

(defun concatenated-misc (stream operation &optional arg1 arg2)
  (let ((left (concatenated-stream-current stream)))
    (when left
      (let* ((current (car left))
	     (misc (sstream-misc current)))
	(case operation
	  (:listen
	   (loop
	     (let ((stuff (funcall misc current :listen)))
	       (cond ((eq stuff :eof)
		      ;; Advance current, and try again.
		      (pop (concatenated-stream-current stream))
		      (setf current
			    (car (concatenated-stream-current stream)))
		      (unless current
			;; No further streams.  EOF.
			(return :eof))
		      (setf misc (sstream-misc current)))
		     (stuff
		      ;; Stuff's available.
		      (return t))
		     (t
		      ;; Nothing available yet.
		      (return nil))))))
	  (t
	   (funcall misc current operation arg1 arg2)))))))

(let*()
(defstruct (string-stream
	     (:include sstream)))
)

(defmethod print-object ((s string-stream) out)
  (print-unreadable-object (s out :type t :identity t)))

(let*()
(defstruct (string-input-stream
	     (:include string-stream
		       (chin #'string-chin)
		       (misc #'string-in-misc))
	     (:constructor %make-string-input-stream (string current end)))
  (string nil :type simple-string)
  (current nil :type fixnum)
  (end nil :type fixnum))
)

(defun string-chin (s)
  (let* ((string (string-input-stream-string s))
	 (current (string-input-stream-current s)))
    (if (< current (string-input-stream-end s))
	(prog1 (schar string current)
	  (setf (string-input-stream-current s) (1+ current))))))

(defun string-in-misc (stream operation &optional arg1 arg2)
  (declare (ignore arg2))
  (case operation
    (:file-position
     (if arg1
	 (setf (string-input-stream-current stream) arg1)
	 (string-input-stream-current stream)))
    (:file-length (length (string-input-stream-string stream)))
    (:unread (decf (string-input-stream-current stream)))
    (:listen (or (/= (the fixnum (string-input-stream-current stream))
		     (the fixnum (string-input-stream-end stream)))
		 :eof))
    (:element-type 'string-char)))

(defun make-string-input-stream (string &optional (start 0 start-p)
					(end nil))
  (let* ((len (length (require-type string 'simple-string))))
    (if start-p
	(unless (and (typep start 'fixnum)
		     (>= start 0)
		     (< start len))
	  (setq start (require-type start `(integer 0 ,len)))))
    (if end
	(unless (and (typep end 'fixnum)
		     (>= end start)
		     (<= end len))
	  (setq end (require-type end `(integer ,start ,len))))
	(setq end len))
    (%make-string-input-stream string start end)))

(defstruct (string-output-stream
	    (:include string-stream
		      (chout #'string-chout)
		      (string-out #'string-sout)
		      (misc #'string-out-misc))
	    (:constructor %%make-string-output-stream (&optional string)))
  (string (make-array 10 :adjustable t :element-type 'base-char :fill-pointer 0) :type string))


(defun %make-string-output-stream (string)
  (unless (and (typep string 'string)
	       (array-has-fill-pointer-p string))
    (error "~S must be a string with a fill pointer."))
  (%%make-string-output-stream string))

(defun make-string-output-stream (&key (element-type 'base-char element-type-p))
  (when (and element-type-p
	     (not (member element-type '(base-character character string-char
					 standard-char))))
    (unless (subtypep element-type 'character)
      (error "~S argument ~S is not a subtype of ~S."
	     :element-type element-type 'character)))
  (%%make-string-output-stream))


(defun string-chout (s char)
  (vector-push-extend char (string-output-stream-string s)))

(defun string-sout (stream string start end)
  (declare (simple-string string) (fixnum start end))
  ;; Could be made faster, certainly ...
  (do* ((dest (string-output-stream-string stream))
	(i start (1+ i)))
       ((= i end) (length dest))
    (declare (fixnum i))
    (vector-push-extend (schar string i) dest)))

(defun string-out-misc (stream operation &optional arg1 arg2)
  (declare (ignore arg2))
  (case operation
    (:file-position
     (if (null arg1)
	 (length (string-output-stream-string stream))))
    (:charpos
     (do* ((string (string-output-stream-string stream))
	   (index (1- (the fixnum (length string)))
		  (1- index))
	   (count 0 (1+ count)))
	  ((< index 0) count)
       (declare (string string)
		(fixnum index count))
       (if (char= (char string index) #\newline)
	   (return count))))
    (:element-type 'string-char)))

;;;One way to indent on newlines:

(let*()
(defstruct (indenting-string-output-stream
	     (:include string-output-stream
		       (chout #'indenting-string-chout)
		       (string-out #'indenting-string-sout))
	     (:constructor make-indenting-string-output-stream (prefixchar indent)))
  (prefixchar nil)
  (indent nil))
)

(defun indenting-string-chout (stream char)
  (let* ((string (string-output-stream-string stream)))
    (vector-push-extend char string)
    (when (eql char #\newline)
      (let* ((indent (or (indenting-string-output-stream-indent stream) 0))
	     (prefixchar (indenting-string-output-stream-prefixchar stream)))
	(declare (fixnum indent))
	(when prefixchar
	    (vector-push-extend prefixchar string)
	    (decf indent))
	(dotimes (i indent)
	  (vector-push-extend #\space string))))))

(defun indenting-string-sout (stream string start end)
  (declare (fixnum start end)
	   (simple-base-string string))
  (do* ((i start (1+ i)))
       ((= i end))
    (declare (fixnum i))
    (indenting-string-chout stream (schar string i))))

(defun get-output-stream-string (s)
  (unless (typep s 'string-output-stream)
    (report-bad-arg s 'string-output-stream))
  (let* ((string (string-output-stream-string s)))
    (prog1 (coerce string 'simple-string)
      (setf (fill-pointer string) 0))))

(defstruct (buffered-sstream (:include sstream
				       (misc #'fd-misc)))
  (fill #'ill-in)
  (flush #'ill-out)
  (in-buffer nil :type (or (simple-array * (*)) nil))
  (in-bufptr nil :type (or macptr nil))
  (out-buffer nil :type (or (simple-array * (*)) nil))
  (out-bufptr nil :type (or macptr nil))
  (element-type (specifier-type 'character) :type ctype)
  (element-size 1 :type fixnum)         ;element size (in octets)
  (lastchar nil :type (or nil character))
  (in-bufsize 0 :type  fixnum)   ;size (in octets) of buffer
  (in-bufidx 0 :type fixnum)    ;index of next element to read
  (in-bufcount 0 :type fixnum)		; -1 if reading & EOF
  (out-bufsize 0 :type fixnum)
  (out-bufidx 0 :type fixnum)
  (out-bufcount 0 :type fixnum)
  (charpos 0 :type (or nil fixnum))  ;position of cursor
  (device -1 :type fixnum)		;file descriptor
  (elements-per-buffer 0 :type fixnum)
  (listen nil)
  (interactive t)
)


(defmethod print-object ((s buffered-sstream) out)
  (print-unreadable-object (s out :type t :identity t)
    (let* ((in-ptr (buffered-sstream-in-bufptr s))
	   (out-ptr (buffered-sstream-in-bufptr s))
	   (direction
	    (if in-ptr
		(if out-ptr :io :input) (if out-ptr :output))))
    (format out "Descriptor ~d direction ~A"
	    (buffered-sstream-device s)
	    direction))))

(defun fd-fill (s)
  (let* ((fd (buffered-sstream-device s))
	 (buf (buffered-sstream-in-bufptr s))
	 (size (buffered-sstream-in-bufsize s)))
    (setf (buffered-sstream-in-bufidx s) 0
	  (buffered-sstream-in-bufcount s) 0)    
    (if (listen s)
	(let* ((n (fd-read fd buf size)))
	  (declare (fixnum n))
	  (if (= n 0)
	      (setf (buffered-sstream-listen s) :eof))
	  (setf (buffered-sstream-in-bufcount s)
		(if (= n 0)
		    -1
		    (the fixnum
		      (* n (the fixnum (buffered-sstream-element-size s)))))))
	(if (eq (buffered-sstream-listen s) :eof)
	    (setf (buffered-sstream-in-bufcount s) -1)))))


(defun fd-flush (s)
  (let* ((fd (buffered-sstream-device s))
	 (buf (buffered-sstream-out-bufptr s))
	 (octets (* (the fixnum (buffered-sstream-element-size s))
		    (the fixnum (buffered-sstream-out-bufidx s)))))
    (declare (fixnum octets))
    (when (> octets 0)
      (setf (buffered-sstream-out-bufidx s) 0)
      (fd-write fd buf octets))
    (setf (buffered-sstream-dirty s) nil)))
	

(defun fd-chout (s ch)
  (let* ((idx (buffered-sstream-out-bufidx s))
	 (limit (buffered-sstream-elements-per-buffer s)))
    (declare (fixnum idx limit))
    (when (= idx limit)
      (funcall (buffered-sstream-flush s) s)
      (setq idx 0))
    (setf (buffered-sstream-out-bufidx s) (the fixnum (1+ idx)))
    (setf (buffered-sstream-dirty s) t)
    (if (eq ch #\newline)
	(setf (buffered-sstream-charpos s) 0)
	(incf (buffered-sstream-charpos s)))
    (setf (schar (buffered-sstream-out-buffer s) idx) ch)))

(defun fd-bout (s b)
  (let* ((idx (buffered-sstream-out-bufidx s))
	 (limit (buffered-sstream-elements-per-buffer s)))
    (declare (fixnum idx limit))
    (when (= idx limit)
      (funcall (buffered-sstream-flush s) s)
      (setq idx 0))
    (setf (buffered-sstream-dirty s) t)
    (setf (buffered-sstream-out-bufidx s) (the fixnum (1+ idx)))
    (setf (uvref (buffered-sstream-out-buffer s) idx) b)))

; start & end are byte (octet) positions
(defun fd-ivector-out (s v start end)
  (declare (fixnum start end))
  (let* ((length (- end start))
	 (written 0)
	 (bufsize (buffered-sstream-out-bufsize s))
	 (buffer (buffered-sstream-out-buffer s)))
    (declare (fixnum length written bufsize))
    (do* ((pos start (+ pos written))
	  (left length (- left written)))
	 ((= left 0))
      (declare (fixnum pos left))
      (let* ((index (buffered-sstream-out-bufidx s))
	     (avail (- bufsize index)))
	(declare (fixnum avail index))
	(cond
	  ((= (setq written avail) 0)
	   (funcall (buffered-sstream-flush s) s))
	  (t
	   (if (> written left)
	       (setq written left))
	   (%copy-ivector-to-ivector v pos buffer index written)
	   (if (= (the fixnum (setf (buffered-sstream-out-bufidx s)
				    (the fixnum (+ index written))))
		  bufsize)
	       (funcall (buffered-sstream-flush s) s))))))))

(defun fd-string-out (s string start end)
  (fd-ivector-out s string start end)
  (let* ((last-newline (if (find #\newline string :start start :end end)
                           (position #\newline string
                                     :from-end t
                                     :start start
                                     :end end))))
    (if last-newline
        (setf (buffered-sstream-charpos s)
              (- end last-newline 1))
        (incf (buffered-sstream-charpos s) (- end start)))))
      
(defun fd-bin (s)
  (setf (buffered-sstream-listen s) nil)
  (loop
      (let* ((pos (buffered-sstream-in-bufidx s))
             (count (buffered-sstream-in-bufcount s)))
        (declare (fixnum pos count))
        (when (< count 0)
          (return))			;EOF: return nil
        (when (< pos count)
          (return
            (prog1 (uvref (buffered-sstream-in-buffer s) pos)
              (setf (buffered-sstream-in-bufidx s) (the fixnum (1+ pos))))))
        (funcall (buffered-sstream-fill s) s))))

(defun fd-chin (s)
  (setf (buffered-sstream-listen s) nil)
  (let* ((ch (buffered-sstream-lastchar s)))
    (if ch
      (progn
	(setf (buffered-sstream-lastchar s) nil))
      (loop
	(let* ((fd (buffered-sstream-device s))
	       (pos (buffered-sstream-in-bufidx s))
	       (count (buffered-sstream-in-bufcount s)))
	  (declare (fixnum pos count))
	  (when (< count 0)
	    (return))			;EOF: return nil
	  (when (< pos count)
	    (setf (buffered-sstream-in-bufidx s) (the fixnum (1+ pos)))
	    (setq ch (schar (buffered-sstream-in-buffer s) pos))
	    (return))
	  (%proxy-wait-for-input fd)
	  (funcall (buffered-sstream-fill s) s))))
    (when (and *dribble-stream* (or (eq s *terminal-input*)
				    (eq s *terminal-io*)))
      (tyo ch *dribble-stream*))
    ch))

(defun fd-misc (stream operation &optional arg1 arg2)
  (declare (ignore arg2))
  (case operation
    (:listen 
     (or (< (buffered-sstream-in-bufidx stream)
	    (buffered-sstream-in-bufcount stream))
	 (buffered-sstream-listen stream)
	 (let* ((avail (fd-input-available-p (buffered-sstream-device stream))))
	   (if (and (not avail) (not (buffered-sstream-interactive stream)))
	       (setq avail :eof))
	   (setf (buffered-sstream-listen stream) avail))))
    (:unread
     (setf (buffered-sstream-lastchar stream) arg1)
     (setf (buffered-sstream-listen stream) t))
    (:close
     (funcall (buffered-sstream-misc stream) stream :finish-output)
     (fd-close (buffered-sstream-device stream))
     (let* ((in-buffer (buffered-sstream-in-buffer stream))
            (out-buffer (buffered-sstream-out-buffer stream)))
       (setf (buffered-sstream-in-buffer stream) nil
	     (buffered-sstream-in-bufptr stream) nil
             (buffered-sstream-out-buffer stream) nil
             (buffered-sstream-out-bufptr stream) nil)
       (if in-buffer (%dispose-heap-ivector in-buffer))
       (if (and out-buffer (not (eq in-buffer out-buffer)))
           (%dispose-heap-ivector out-buffer))))
    (:clear-input
     (setf (buffered-sstream-lastchar stream) nil)
     (setf (buffered-sstream-in-bufidx stream) 0)
     (setf (buffered-sstream-in-bufcount stream) 0)
     #|
     (catch 'eof-input-catcher
       (loop
	(let ((count (unix:unix-select (1+ (fd-stream-fd stream))
				       (the (unsigned-byte 32)
					    (ash 1 (fd-stream-fd stream)))
				       0 0 0)))
	  (cond ((eql count 1)
		 (do-input stream)
		 (setf (fd-stream-ibuf-head stream) 0)
		 (setf (fd-stream-ibuf-tail stream) 0))
		(t
		 (return t))))))
     |#)
    ((:force-output :finish-output)
     (let* ((flush (buffered-sstream-flush stream)))
       (unless (eq flush #'ill-out)
	 (funcall flush stream))))
    
    (:element-type
     (buffered-sstream-element-type stream))
    (:interactive-p
     (buffered-sstream-interactive stream))
    (:line-length
     80)
    (:charpos
     (buffered-sstream-charpos stream))))

; Could lapify ...
(defun %octets-to-elements (n element-size)
  (if (eql 8 element-size)
    n
    (values (floor (* n 8) element-size))))

; Could lapify.
(defun %elements-to-octets (n element-size)
  (if (eql 8 element-size)
    n
    (values (ceiling (* n element-size) 8))))

#+ppc-target
(defppclapfunction %vect-data-to-macptr ((vect arg_y) (ptr arg_z))
  ; put address of vect data in macptr
  (addi arg_y arg_y arch::misc-data-offset)
  (stw arg_y arch::macptr.address arg_z)
  (blr))

#+sparc-target
(defsparclapfunction %vect-data-to-macptr ((vect %arg_y) (ptr %arg_z))
  ; put address of vect data in macptr
  (inc arch::misc-data-offset vect)
  (retl)
   (st vect (ptr arch::macptr.address)))


(defloadvar *heap-ivectors* ())

(defun %make-heap-ivector (subtype size-in-bytes &optional size-in-elts)
  (if (not (fixnump size-in-elts))(error "Need size in elts = fixnum")) ; not really optional, is smaller than fixnum
  (with-macptrs ((ptr (malloc (+ size-in-bytes (+ 4 2 7))))) ; 4 for header, 2 for delta, 7 for round up
    (let ((vect (fudge-heap-pointer ptr subtype size-in-elts))
          (p (%null-ptr)))
      (%vect-data-to-macptr vect p)
      (push vect *heap-ivectors*)
      (values vect p))))

#+ppc-target
(defppclapfunction fudge-heap-pointer ((ptr arg_x) (subtype arg_y) (len arg_z))
  (check-nargs 3)
  (macptr-ptr imm1 ptr) ; address in macptr
  (addi imm0 imm1 9)     ; 2 for delta + 7 for alignment
  (rlwinm imm0 imm0 0 0 28)   ; Clear low three bits to align
  (subf imm1 imm1 imm0)  ; imm1 = delta
  (sth imm1 -2 imm0)     ; save delta halfword
  (unbox-fixnum imm1 subtype)  ; subtype at low end of imm1
  (rlwimi imm1 len (- arch::num-subtag-bits arch::fixnum-shift) 0 (- 31 arch::num-subtag-bits))
  (stw imm1 0 imm0)       ; store subtype & length
  (addi arg_z imm0 arch::fulltag-misc) ; tag it, return it
  (blr))

#+sparc-target
(defsparclapfunction fudge-heap-pointer ((ptr %arg_x) (subtype %arg_y) (len %arg_z))
  (check-nargs 3)
  (macptr-ptr ptr %imm1) ; address in macptr
  (add %imm1 9 %imm0)     ; 2 for delta + 7 for alignment
  (andn %imm0 7 %imm0)   ; Clear low three bits to align
  (sub %imm0 %imm1 %imm1)  ; imm1 = delta
  (sth %imm1 (%imm0 -2))     ; save delta halfword
  (unbox-fixnum subtype %imm2)  ; subtype at low end of imm2
  (sll len  (- arch::num-subtag-bits arch::fixnum-shift) %imm1)
  (or %imm2 %imm1 %imm1)
  (st %imm1 (%imm0))	  ; store header
  (retl)
    (add %imm0 arch::fulltag-misc %arg_z))  ; tag it, return it

(defun %dispose-heap-ivector (v)
  (if  (uvectorp v) ;(%heap-ivector-p v)
    (with-macptrs (p)
      (setq *heap-ivectors* (delq v *heap-ivectors*))
      (%%make-disposable p v)
      (free p))))

(defun make-heap-buffer (element-type element-count)
  (let* ((subtype arch::subtag-simple-base-string)
         (size-in-octets element-count))
    (unless (subtypep element-type 'character)
      (case element-type
        (unsigned-byte (setq element-type '(unsigned-byte 8)))
        (signed-byte (setq element-type '(signed-byte 8))))
      (let* ((signed (list 'signed-byte 0))
             (unsigned (list 'unsigned-byte 0)))
        (declare (dynamic-extent signed unsigned))
        (do* ((i 8 (+ i 8))
              (octets 1 (1+ octets))
              (match nil))             
             ((> i 32) (report-bad-arg element-type '(or character
                                                      (unsigned-byte 32)
                                                      (signed-byte 32))))
          (setf (cadr signed) i)
          (setf (cadr unsigned) i)
          (if (subtypep element-type signed)
              (setq match (copy-list signed))
              (if (subtypep element-type unsigned)
                  (setq match (copy-list unsigned))))
          (when match
            (setq size-in-octets (* octets element-count)
                  subtype (element-type-subtype match)
                  element-type match)
            (return)))))
    (multiple-value-bind (buf p)
        (%make-heap-ivector subtype element-count size-in-octets)
      (values buf p element-type))))

#+ppc-target          
(defppclapfunction %%make-disposable ((ptr arg_y) (vector arg_z))
  (check-nargs 2)
  (subi imm0 vector arch::fulltag-misc) ; imm0 is addr = vect less tag
  (lhz imm1 -2 imm0)   ; get delta
  (sub imm0 imm0 imm1)  ; vector addr (less tag)  - delta is orig addr
  (stw imm0 arch::macptr.address ptr) 
  (blr))

#+sparc-target
(defsparclapfunction %%make-disposable ((ptr %arg_y) (vector %arg_z))
  (check-nargs 2)
  (sub vector arch::fulltag-misc %imm0)
  (lduh (%imm0 -2) %imm1)
  (sub %imm0 %imm1 %imm0)
  (retl)
   (st %imm0 (ptr arch::macptr.address)))


(defvar *elements-per-buffer* 2048)  ; default buffer size for file io

(defun make-tty-stream (fd &key
			   (direction :input)
                           (interactive t)
			   (elements-per-buffer *elements-per-buffer*))
  (let* ((in-p (member direction '(:io :input)))
         (out-p (member direction '(:io :output))))
  (multiple-value-bind (in-buffer in-ptr)
      (if in-p
          (make-heap-buffer 'character elements-per-buffer))
    (multiple-value-bind (out-buffer out-ptr)
        (if out-p
            (make-heap-buffer 'character elements-per-buffer))
      (let* ((s (make-buffered-sstream
                 :in-buffer in-buffer
                 :in-bufptr in-ptr
                 :out-buffer out-buffer
                 :out-bufptr out-ptr
                 :element-type 'character
                 :device fd
                 :elements-per-buffer elements-per-buffer
                 :interactive interactive)))
        (when in-p
          (setf (buffered-sstream-chin s) #'fd-chin
                (buffered-sstream-fill s) #'fd-fill
                (buffered-sstream-in-bufsize s) elements-per-buffer))
        (when out-p
          (setf (buffered-sstream-chout s) #'fd-chout
                (buffered-sstream-string-out s) #'fd-string-out
                (buffered-sstream-flush s) #'fd-flush
                (buffered-sstream-charpos s) 0
                (buffered-sstream-out-bufsize s) elements-per-buffer))
        s)))))
          
			       
  
  
			   
(defmethod initialize-instance ((v instance-initialize-mixin) &rest initargs
                                &key ignore)
  (declare (dynamic-extent initargs))
  (declare (ignore ignore))
  (call-next-method)
  (apply #'instance-initialize v initargs))

; Make sure a primary method exists.
(defmethod instance-initialize ((v instance-initialize-mixin) &key))

(def-aux-init-functions instance-initialize-mixin #'instance-initialize)

(defclass clos-stream (instance-initialize-mixin)
  ((direction :initarg :direction :initform nil :reader stream-direction)))

(defparameter *stream-class* (find-class 'clos-stream))

(defun streamp (thing)
  (or (sstreamp thing)
      (not (null (memq *stream-class* (%inited-class-cpl (class-of thing)))))))

(defclass input-stream (clos-stream) ())
(defclass output-stream (clos-stream) ())

(defmethod instance-initialize :after ((stream input-stream) &key)
  (let ((direction (slot-value stream 'direction)))
    (if (null direction)
      (set-slot-value stream 'direction :input)
      (if (eq direction :output)
        (set-slot-value stream 'direction :io)))))

(defmethod instance-initialize :after ((stream output-stream) &key)
  (let ((direction (slot-value stream 'direction)))
    (if (null direction)
      (set-slot-value stream 'direction :output)
      (if (eq direction :input)
        (set-slot-value stream 'direction :io)))))

(defmethod stream-tyo ((stream clos-stream) char)
  (declare (ignore char))
  (error "stream ~S is not capable of output" stream))

(defun stream-write-entire-string (stream string)
  (write-string string stream :start 0 :end (length string)))

; The only caller is WRITE-STRING, which does sanity checking
; and dereferences the string.
(defmethod stream-write-string ((stream clos-stream) string start end)
  (declare (simple-base-string string)
	   (fixnum start end))
  (do* ((i start (1+ i)))
       ((= i end))
    (declare (fixnum i))
    (tyo (schar string i) stream)))

(defmethod stream-tyi (stream)
  (error "~s is not capable of input" stream))


(defmethod stream-untyi ((stream clos-stream) char)
  (declare (ignore char))
  (error "stream ~S is not capable of input" stream))

(defun terminator-predicate (terminator)
  (cond ((characterp terminator) #'eq)
        ((stringp terminator)
         #'%str-member)
        ((functionp terminator)
         #'(lambda (char terminator) (funcall terminator char)))))

(defmethod stream-read-string ((stream clos-stream) string start end terminator)
  (setq start (require-type start 'fixnum))
  (setq end (require-type end 'fixnum))
  (let ((pred (terminator-predicate terminator))
        (start start)
        (end end))
    (declare (fixnum start end))
    (while (< start end)
      (let ((char (tyi stream)))
        (when (and pred (funcall pred char terminator))
          (untyi char stream)
          (return-from stream-read-string start))
        (setf (schar string start) char)
        (incf start))))
  end)


(defmethod stream-eofp ((stream clos-stream))
  (error "stream ~S is not capable of input" stream))

(defmethod stream-force-output ((stream output-stream)) nil)

(defmethod stream-close ((stream clos-stream))
  (set-slot-value stream 'direction :closed)
  t)

(defun open-stream-p (stream)
  (setq stream (require-type stream 'clos-stream))
  (neq (stream-direction stream) :closed))

(defmethod stream-fresh-line ((stream output-stream))
  (terpri stream)
  t)

(defmethod stream-line-length ((stream clos-stream))
  "This is meant to be shadowed by particular kinds of streams,
   esp those associated with windows."
  80)

(defmethod interactive-stream-p ((stream clos-stream)) nil)

(defmethod stream-clear-input ((stream input-stream)) nil)

(defmethod stream-listen ((stream input-stream))
  (not (eofp stream)))

(defmethod stream-abort ((stream clos-stream))
  (set-slot-value stream 'direction :closed))

(defmethod stream-filename ((stream clos-stream)) nil)



(defclass two-way-stream (input-stream output-stream)
  ((input-stream :initarg :input-stream :reader two-way-stream-input-stream)
   (output-stream :initarg :output-stream :reader two-way-stream-output-stream)))

(defun make-two-way-stream (input-stream output-stream)
  (make-instance 'two-way-stream
                 :input-stream input-stream :output-stream output-stream))

(defmethod stream-tyo ((stream two-way-stream) char)
  (tyo char (slot-value stream 'output-stream)))

(defvar *two-way-stream-class* (find-class 'two-way-stream))

(defmethod stream-column ((stream two-way-stream))
  (column (slot-value stream 'output-stream)))


(defmethod stream-tyi ((stream two-way-stream))
  (force-output (slot-value stream 'output-stream))
  (tyi (slot-value stream 'input-stream)))


(defmethod stream-untyi ((stream two-way-stream) char)
  (untyi char (slot-value stream 'input-stream)))

(defmethod stream-eofp ((stream two-way-stream))
  (eofp (slot-value stream 'input-stream)))

(defmethod stream-force-output ((stream two-way-stream))
  (force-output (slot-value stream 'output-stream)))

(defmethod stream-finish-output ((stream two-way-stream))
  (finish-output (slot-value stream 'output-stream)))

(defmethod stream-close :after ((stream two-way-stream))
  (slot-makunbound stream 'input-stream)
  (slot-makunbound stream 'output-stream))

(defmethod stream-fresh-line ((stream two-way-stream))
  (fresh-line (slot-value stream 'output-stream)))

(defmethod stream-clear-input ((stream two-way-stream))
  (clear-input (slot-value stream 'input-stream)))

(defmethod stream-listen ((stream two-way-stream))
  (listen (slot-value stream 'input-stream)))

(defclass echoing-two-way-stream (two-way-stream)
  ())

(defmethod stream-tyi ((s echoing-two-way-stream))
  (let* ((out (slot-value s 'output-stream)))
    (force-output out)
    (let* ((ch (tyi (slot-value s 'input-stream))))
      (when ch
	(when (typep out 'buffered-sstream)
	  (if (eq ch #\newline)
	      (setf (buffered-sstream-charpos out) 0)
	      (incf (buffered-sstream-charpos out)))))
      ch)))

(defloadvar *fd-set-size*
    (ff-call (%kernel-import arch::kernel-import-fd-setsize-bytes)
             :unsigned-fullword))

(defun old-fd-input-available-p (fd)
  (rlet ((arg :long))
   (%set-long arg 0)
   (when (zerop
          (ff-call (%kernel-import arch::kernel-import-ioctl)
                        :unsigned-fullword fd
                        :unsigned-fullword FIONREAD
                        :address arg
                        :signed-fullword))
    (> (%get-long arg) 0))))

(defun fd-zero (fdset)
  (ff-call (%kernel-import arch::kernel-import-do-fd-zero)
	   :address fdset
	   :void))

(defun fd-set (fd fdset)
  (ff-call (%kernel-import arch::kernel-import-do-fd-set)
	   :unsigned-fullword fd
	   :address fdset
	   :void))

(defun fd-clr (fd fdset)
  (ff-call (%kernel-import arch::kernel-import-do-fd-clr)
	   :unsigned-fullword fd
	   :address fdset
	   :void))

(defun fd-is-set (fd fdset)
  (not (= 0 (the fixnum (ff-call (%kernel-import arch::kernel-import-do-fd-is-set)
				 :unsigned-fullword fd
				 :address fdset
				 :unsigned-fullword)))))

(defun fd-input-available-p (fd)
  (rletZ ((tv :timeval))
    (%stack-block ((infds *fd-set-size*))
      (fd-zero infds)
      (fd-set fd infds)
      (let* ((res (external-call "select"
                                 :signed-fullword (1+ fd)
                                 :address infds
                                 :address (%null-ptr)
                                 :address (%null-ptr)
                                 :address tv
                                 :signed-fullword)))
        (> res 0)))))


; Some simple explicit storage management for cons cells
(defvar *cons-pool* (%cons-pool nil))

(defun cheap-cons (car cdr)
  (let* ((pool *cons-pool*)
         (cons (pool.data pool)))
    (if cons
      (locally (declare (type cons cons))
        (setf (pool.data pool) (cdr cons)
              (car cons) car
              (cdr cons) cdr)
        cons)
      (cons car cdr))))

(defun free-cons (cons)
  (when (consp cons)
    (locally (declare (type cons cons))
      (setf (car cons) nil
            (cdr cons) nil)
      (let* ((pool *cons-pool*)
             (freelist (pool.data pool)))
        (setf (pool.data pool) cons
              (cdr cons) freelist)))))

(defun cheap-copy-list (list)
  (let ((l list)
        res)
    (loop
      (when (atom l)
        (return (nreconc res l)))
      (setq res (cheap-cons (pop l) res)))))

(defun cheap-list (&rest args)
  (declare (dynamic-extent args))
  (cheap-copy-list args))

; Works for dotted lists
(defun cheap-free-list (list)
  (let ((l list)
        next-l)
    (loop
      (setq next-l (cdr l))
      (free-cons l)
      (when (atom (setq l next-l))
        (return)))))

(defmacro pop-and-free (place)
  (setq place (require-type place 'symbol))     ; all I need for now.
  (let ((list (gensym))
        (cdr (gensym)))
    `(let* ((,list ,place)
            (,cdr (cdr ,list)))
       (prog1
         (car ,list)
         (setf ,place ,cdr)
         (free-cons ,list)))))

; Support for defresource & using-resource macros
(defun make-resource (constructor &key destructor initializer)
  (%cons-resource constructor destructor initializer))

(defun allocate-resource (resource)
  (setq resource (require-type resource 'resource))
  (let ((pool (resource.pool resource))
        res)
    (without-interrupts
     (let ((data (pool.data pool)))
       (when data
         (setf res (car data)
               (pool.data pool) (cdr (the cons data)))
         (free-cons data))))
    (if res
      (let ((initializer (resource.initializer resource)))
        (when initializer
          (funcall initializer res)))
      (setq res (funcall (resource.constructor resource))))
    res))

(defun free-resource (resource instance)
  (setq resource (require-type resource 'resource))
  (let ((pool (resource.pool resource))
        (destructor (resource.destructor resource)))
    (when destructor
      (funcall destructor instance))
    (without-interrupts
     (setf (pool.data pool)
           (cheap-cons instance (pool.data pool)))))
  resource)      

(defresource *string-output-stream-pool*
  :constructor (make-instance 'string-output-stream)
  :initializer 'stream-clear-output)



;;;File streams.

; The idea that EXTERNAL-FORMAT has anything to do with mac file types is, er, um,
; "not well developed."
(defun open (filename &key (direction :input)
                      (element-type 'base-char)
                      (if-exists :error)
                      (if-does-not-exist (if (or (eq direction :input)
                                                 ;(eq if-exists :overwrite)
                                                 (eq if-exists :append))
                                           :error :create))
                      (external-format :default)
                      (elements-per-buffer *elements-per-buffer*)		          
)  
  (loop
    (restart-case
      (return (open-internal
               filename direction element-type if-exists if-does-not-exist
               external-format elements-per-buffer 
               ))
      (retry-open ()
                  :report (lambda (stream) (format stream "Retry opening ~s" filename))
                  nil))))

(defun scriptruns-p (truename)
  (declare (ignore truename))
  nil)



(defun open-internal (filename direction element-type if-exists if-does-not-exist
                                    external-format  elements-per-buffer
                                    &aux fstream truename
		                    tem-file kind)
  (declare (ignore external-format))
  (block open
    (if (or (memq element-type '(:default character base-char))
            (subtypep element-type 'character))
      (if (eq element-type :default)(setq element-type 'base-char))           ; should character output be prohibited to non-TEXT files?  if so, why ?
      (progn
        (setq element-type (type-expand element-type))
        (cond ((equal element-type '#.(type-expand 'signed-byte))
               (setq element-type '(signed-byte 8)))
              ((equal element-type '#.(type-expand 'unsigned-byte))
               (setq element-type '(unsigned-byte 8))))))
    (case direction
      (:probe (setq if-exists :ignored if-does-not-exist nil))
      (:input (setq if-exists :ignored))
      ((:io :output :shared) nil)
      (t (report-bad-arg direction '(member :input :output :io :probe :shared))))
    (multiple-value-setq (truename kind)(probe-file-x filename))
    (if truename
      (if (eq kind :directory)
        (if (eq direction :probe)
          (return-from open nil)
          (signal-file-error $xdirnotfile truename))
        (if (setq filename (if-exists if-exists truename "Open ..."))
          (progn
            (multiple-value-setq (truename kind) (probe-file-x filename))
            (cond 
             ((not truename)
              (setq truename (create-file filename)))
             ((memq direction '(:output :io :shared))
#|                      
              ; this prevents us from writing a file that is open for anything            
              ; but does not protect against reading a file that is open for :output
              (when (and bits (eq direction :output)(neq 0 (logand bits #x81)))
                (signal-file-error $xfilebusy truename))
|#
              (when (eq if-exists :supersede)
                ;;; FIX this: rename existing file, create new one
                (create-file truename )         ; for now, clobber existing file.
                ))))
          (return-from open nil)))
      (if (setq filename (if-does-not-exist if-does-not-exist filename))
        (setq truename (create-file filename))
        (return-from open nil)))
    (setq fstream (make-instance (getf 
                                  (if (memq element-type '(character base-char extended-char standard-char))
                                    '(:input input-file-stream
                                      :output output-file-stream
                                      :io io-file-stream
                                      :shared io-file-stream
                                      :probe file-stream)
                                    '(:input input-binary-file-stream
                                      :output output-binary-file-stream
                                      :io io-binary-file-stream
                                      :shared io-binary-file-stream
                                      :probe binary-file-stream))
                                  direction)
                    :filename truename
                    :element-type element-type
                    :direction direction
                    :elements-per-buffer elements-per-buffer
                    :orig-filename tem-file))
    (cond ((eq if-exists :append)
           (file-position fstream :end))
          ((and (memq direction '(:io :output :shared)) (neq if-exists :overwrite))
           (file-length fstream 0)))   
    fstream))


(defun gen-file-name (path)
  (let* ((date (mac-file-write-date path))
         (tem-path (merge-pathnames (make-pathname :name (%integer-to-string date) :type "tem" :defaults nil) path)))
    (loop
      (when (not (probe-file tem-path)) (return tem-path))
      (setf (%pathname-name tem-path) (%integer-to-string (setq date (1+ date)))))))




(defun probe-file-x (path)
  (%probe-file-x (namestring (translate-logical-pathname path))))


(defclass file-stream (clos-stream)
  ((my-file-name :initarg :filename :reader stream-filename)
   (column.fblock :initform (cons 0 nil) :reader column.fblock)
   (fblock)
   (element-type :initform 'base-char :initarg :element-type)
   (orig-file-name :initarg :orig-filename)))

(defclass input-file-stream (file-stream input-stream) ())

(defclass output-file-stream (file-stream output-stream) ())

(defclass io-file-stream (input-file-stream output-file-stream) ())

(defmethod print-object ((stream file-stream) output-stream)
  (print-unreadable-object (stream output-stream)
    (format output-stream "~A ~S to ~S"
            (or (stream-direction stream) :probe)
            (class-name (class-of stream))
            (namestring (stream-filename stream)))))

(defmethod instance-initialize :after ((stream file-stream) &key 
                                       direction 
                                       (elements-per-buffer *elements-per-buffer*)
                                       (element-type 'base-char))
  (setf (slot-value stream 'direction) direction)
  (when direction
    (let* ((filename (slot-value stream 'my-file-name))
           (fblock (%fopen  filename
                           (case direction
                             ((:input :probe) O_RDONLY)
                             (:output O_RDWR)
                             (:io O_RDWR)
                             (:shared O_RDWR))
                           nil
                           stream
                           element-type
                           elements-per-buffer
                           nil)))
      (set-slot-value stream 'fblock fblock)
      (setf (cdr (column.fblock stream)) fblock)
      (push stream *open-file-streams*))))

(defmethod stream-element-type ((stream file-stream))
  (slot-value stream 'element-type))

(defmethod stream-tyi ((stream input-file-stream))
  (%ftyi (slot-value stream 'fblock)))


(defvar *input-file-stream-class* (find-class 'input-file-stream))



(defmethod stream-untyi ((stream input-file-stream) char)
  (%funtyi (slot-value stream 'fblock) char))

(defmethod stream-eofp ((stream input-file-stream))
  (%feofp (slot-value stream 'fblock)))

(defmethod stream-close :after ((stream file-stream))
  (when (slot-boundp stream 'fblock)
    (let ((fblock (slot-value stream 'fblock)))
      (when fblock (%fclose fblock))
      (setq *open-file-streams* (nremove stream *open-file-streams*))
      (let ((tem (slot-value stream 'orig-file-name))) ; orig-file-name s.b. called orig-contents 
        (when tem (unlock-file tem)(delete-file tem)))
      (slot-makunbound stream 'fblock)
      (slot-makunbound stream 'column.fblock)
      t)))

(defmethod stream-abort ((stream file-stream))
  (when (slot-boundp stream 'fblock)
    (let ((fblock (slot-value stream 'fblock)))
      (when fblock (%fclose fblock))
      (setq *open-file-streams* (nremove stream *open-file-streams*))
      (let ((tem (slot-value stream 'orig-file-name)))
        (when tem          
          (let ((new (slot-value stream 'my-file-name)))
            (unlock-file tem) ; without interrupts ?
	    (delete-file new)
	    (rename-file tem new))))
      (slot-makunbound stream 'fblock)
      t)))

(defmethod stream-tyo ((stream output-file-stream) char)
  (file-stream-tyo (column.fblock stream) char))

(defun file-stream-tyo (column.fblock char)
  (let ((col (car column.fblock)))
    (%rplaca column.fblock
          (if (eq char #\return)
            0
            (if col (1+ col))))
    (%ftyo (%cdr column.fblock) char)))

(defmethod stream-fresh-line ((stream output-file-stream))
  (unless (eql 0 (column stream))
    (tyo #\newline stream)
    t))

(defvar *output-file-stream-class* (find-class 'output-file-stream))

(defmethod stream-column ((stream file-stream))
  (let* ((column.fblock (column.fblock stream))
         (direction (slot-value stream 'direction))
         column)
    (if (and (eq direction :output)
             (setq column (car column.fblock)))
      column
      (let* ((fblock (cdr column.fblock))
             (old-pos (%fpos fblock nil)))
        (unwind-protect
          (setq column
                (do* ((pos (1- old-pos) (1- pos)))
                     ((<= pos 0) old-pos)
                  (%fpos fblock pos)
                  (if (eq #\return (%ftyi fblock))
                    (return (- old-pos pos 1)))))
          (%fpos fblock old-pos))
        (if (eq direction :output)
          (setf (car column.fblock) column))))))

(defmethod clos-file-length ((stream file-stream) &optional new-length)
  (let* ((block (slot-value stream 'fblock))
         (old-size (%fsize block)))
    (when (and new-length (< new-length old-size))
      (%fpos block new-length t))
    (when new-length (%fsize block new-length))
    (or new-length old-size)))

(defun file-length (stream &optional new-length)
  (etypecase stream
    (sstream (funcall (sstream-misc stream) stream :file-length new-length))
    (clos-stream (clos-file-length stream new-length))))

(defun file-position (stream &optional position)
  (etypecase stream
    (sstream (funcall (sstream-misc stream) stream :file-position position))
    (clos-stream (stream-position stream position))))

(defmethod stream-position ((stream file-stream) &optional position)
  (let* ((column.fblock (column.fblock stream))
         (fblock (cdr column.fblock)))
    (when (null position)
      (return-from stream-position (%fpos fblock nil)))
    (let ((length (file-length stream)))
      (setq position (cond ((eq position :start) 0)
                           ((eq position :end) length)
                           (t (require-type position 'integer))))
      (when (> position length)
        (if (typep stream 'output-file-stream)
          ; Why not do something useful?
          (let* ((diff (- position length)))
            (file-length stream position)
            (file-position stream length)
            (if (typep (stream-element-type stream) 'character)
              (dotimes (i diff) (write-char #\null stream))
              (dotimes (i diff) (write-byte 0 stream))))
          (error "~s is beyond the length of the file, which is currently ~s"
                 position length))))
    (when (eq (slot-value stream 'direction) :output)
      ; below says I dunno what the column is
      (setf (car column.fblock)(if (eq 0 position) 0 nil)))
    (%fpos fblock position)
    t))

(defmethod stream-force-output ((stream output-file-stream))
  (%fclose (slot-value stream 'fblock) t))

(defmethod stream-finish-output ((stream output-file-stream))
  (%fclose (slot-value stream 'fblock) 'finish-output))




;; Kindler Gentler World section

(defmethod stream-listen (clos-stream)
  (declare (ignore clos-stream))
  nil)


;;; Initialize the global streams

; These are defparameters because they replace the ones that were in l1-init
; while bootstrapping.


(defparameter *terminal-io* nil)
(defparameter *debug-io* nil)
(defparameter *query-io* nil)
(defparameter *error-output* nil)
(defparameter *standard-input* nil)
(defparameter *standard-output* nil)
(defparameter *trace-output* nil)



(proclaim '(stream 
          *query-io* *debug-io* *error-output* *standard-input* 
          *standard-output* *trace-output*))


; end of L1-streams.lisp

#|
	Change History (most recent last):
	2	12/27/94	akh	merge with d13
|# ;(do not edit past this line!!)
