
;;;-*-Mode: LISP; Package: CCL -*-

;;	Change History (most recent first):

(defstruct (file-sstream
             (:include buffered-sstream
		       (interactive nil)
                       (misc #'file-stream-misc))
             (:constructor %%make-file-sstream ))
  pathname
  original-name
  delete-original
  (filepos 0)
  (filesize 0)
  (direction :input))

(defmethod print-object ((s file-sstream) out)
  (print-unreadable-object (s out :type t :identity t)
    (let* ((dir (file-sstream-direction s))
	   (fd (file-sstream-device s)))
      (if (member dir '(:input :io))
	  (write-char #\< out))
      (write-char #\- out)
      (if (member dir '(:output :io))
	  (write-char #\> out))
      (format out " ~s" (file-sstream-pathname s))
      (when fd (format out " [~d]" fd)))))


(defun file-sstream-position (s newpos))

(defun file-stream-misc (s op &optional arg1 arg2)
  (let* ((fd (file-sstream-device s)))
    (case op
      (:file-position
       (file-sstream-position s arg1))
      (:file-size
       (file-sstream-size s arg1))
      (t
       (fd-misc s op arg1 arg2)))))

;; File exists (or at least has been created already if
;; opening for output.)
(defun make-file-sstream (namestring &key
                                     pathname 
                                     (direction :input)
                                     (element-type 'character)
                                     (elements-per-buffer *elements-per-buffer*)
                                     (original-name nil)
                                     (delete-original nil))
  (multiple-value-bind (fd errno)
      (fd-open namestring (ecase direction
                            (:input O_RDONLY)
                            (:output O_WRONLY)
                            (:io O_RDWR)))
    (if errno (signal-posix-error errno))
    (multiple-value-bind (buffer ptr element-type)
        (make-heap-buffer element-type elements-per-buffer)
      (let* ((octets (subtag-bytes (typecode buffer) elements-per-buffer))
             (s (%%make-file-sstream
                 :pathname pathname
                 :original-name original-name
                 :delete-original delete-original
                 :device fd
                 :direction direction
                 :buffer buffer
                 :bufptr ptr
                 :element-type element-type
                 :element-size (floor octets elements-per-buffer)
                 :bufsize octets
                 :elements-per-buffer elements-per-buffer
		 :filesize (fd-size fd)
                 )))
        (when (member direction '(:input :io))
          (setf (file-sstream-fill s) #'fd-fill)
          (case element-type
            (character (setf (file-sstream-chin s) #'fd-chin))
            (t (setf (file-sstream-bin s) #'fd-bin))))
        (when (member direction '(:output :io))
          (setf (file-sstream-flush s) #'fd-flush)
          (case element-type
            (character
             (setf (file-sstream-chout s) #'fd-chout)
             (setf (file-sstream-string-out s) #'fd-string-out))
            (t
             (setf (file-sstream-bout s) #'fd-bout))))
        s))))
      
                        
                                    


;structure of fblock

(def-accessors (fblock) %svref
  nil                                   ; 'fblock
  fblock.fd                             ; a file descriptor; nil if closed.
  fblock.lastchar                       ; untyi char or nil
  fblock.dirty                          ; non-nil when dirty
  fblock.buffer                         ; macptr to buffer; nil when closed
  fblock.bufvec                         ; buffer vector; nil when closed
  fblock.bufsize                        ; size (in 8-bit bytes) of buffer
  fblock.bufidx                         ; index of next element to read/write
  fblock.bufcount                       ; # of elements in buffer
  fblock.filepos                        ; 8-bit position at last read/write
  fblock.fileeof                        ; file's logical eof.
  fblock.stream                         ; backptr to file stream
  fblock.element-type                   ; typespec
  fblock.nbits-per-element              ; # of bits per element
  fblock.elements-per-buffer            ; 512 or whatever
  fblock.minval                         ; minimum value of element type or nil: < 0 
  fblock.maxval                         ; maximum value or nil
  fblock.element-bit-offset             ; for non-arefable n-bit elements
  fblock.scriptruns			; goo for files containing fat scripts
  fblock.charpos			; character position as opposed to byte position
  fblock.endrun				; end of current run (beware untyi!)
  fblock.scriptidx			; current index into scriptruns
  fblock.chartable			; char-byte-table of current script
)





    


; Open the file identified by NAME string & VREFNUM.
; MODE is O_RDONLY (read permission) or O_RDWR (read/write permission)
; RESOURCE is non-NIL to open the resource fork instead of the data fork
(defun %fopen (name-or-fd &optional 
                          (mode 2) 
                          resource
                          stream
                          (element-type 'base-char)
                          (elements-per-buffer *elements-per-buffer*)
                          scriptruns
                          (seekable t))
  (let* ((eof nil)
         (subtype nil)
         (element-size nil)
         (min nil)
         (max nil)
         (bit-offset nil)
         (buffer-octet-size  elements-per-buffer))
                                        ; need ppc types here too. ok.
    (let ((maybe-subtype (element-type-subtype element-type)))
      (if (subtypep element-type 'character) ; not really cltl2 but what the foo. Nobody wants 16 bit char files.
          (setq element-size 8 subtype (element-type-subtype 'base-char)) ; $v_sstr)
          (let* ((signed (list 'signed-byte 0))
                 (unsigned (list 'unsigned-byte 0)))
            (declare (dynamic-extent signed unsigned))
            (dotimes (i 32 (report-bad-arg element-type '(or character (signed-byte 32) (unsigned-byte 32))))
              (let* ((j (1+ i)))
                (declare (fixnum j))
                (rplaca (cdr signed) j)
                (unless (= j 1)
                  (if (subtypep element-type signed)
                      (setq element-size j
                            subtype (if (memq j '(8 16 32)) maybe-subtype))))
                (rplaca (cdr unsigned) j)
                (if (subtypep element-type unsigned)
                    (setq element-size j
                          min 0
                          subtype (if (memq j '(1 8 16 32)) maybe-subtype))))
              (if element-size (return)))
            (if min                     ; was unsigned
                (setq max (1- (ash 1 element-size)))
                (setq min (- (ash 1 (1- element-size)))
                      max (1- (ash 1 (1- element-size)))))
            (do* ((m (ash elements-per-buffer 3) (+ m 8)))
                 ((eq element-size (gcd m element-size)) (setq buffer-octet-size (ash m -3)))
              (declare (fixnum m)))
            (unless subtype (setq bit-offset 0)))))
    (setq elements-per-buffer (%octets-to-elements buffer-octet-size element-size))
                                        ;(print (list elements-per-buffer buffer-octet-size))    
    (let* ((fd nil))
      (if (typep name-or-fd 'fixnum)
          (setq fd name-or-fd)
          (let* ((res (fd-open name-or-fd (logior (if resource (ash 1 4) 0) mode))))
            (declare (fixnum res))
            (unless (>= res 0)
              (signal-file-error res name-or-fd))
            (setq fd res)))
      (setq eof (if seekable (fd-size fd)))
      (let* ((bufvector nil)
             (buffer nil))
        (if subtype
            (multiple-value-setq (bufvector buffer)
              (%make-heap-ivector subtype
                                  buffer-octet-size
                                  elements-per-buffer))
            (setq buffer (malloc buffer-octet-size)))
        (let* ((fblock (%istruct
                        'fblock
                        fd
                        nil
                        nil
                        buffer
                        bufvector
                        buffer-octet-size
                        0
                        0
                        (if seekable 0)
                        (if eof (%octets-to-elements eof element-size))
                        stream
                        element-type
                        element-size
                        elements-per-buffer
                        min
                        max
                        bit-offset
                        scriptruns
                        0
                        0 
                        -1
                        nil)))
          (when (> eof 0)
            (%fbRead fblock 0))
          fblock)))))

(defun %fblock.fd (f &optional ok-if-closed)
  (or (fblock.fd f) ok-if-closed (error "~s is closed" (fblock.stream f))))

(eval-when (:execute :compile-toplevel)
  (defmacro with-file-descriptor ((fblock fd &optional ignore-if-closed) &body body)
    `(let* ((,fd (%fblock.fd  ,fblock ,ignore-if-closed)))
       (when (typep ,fd 'fixnum)
         (locally
           (declare (fixnum ,fd))
           ,@body)))))


;(%fclose fblock);
;Returns nil
;Close file referenced by fblock

(defun %fclose (fblock &optional flush-only)
  (with-file-descriptor (fblock fd t)
    (%fbFlush fblock)
    (unless flush-only
      (errchk (fd-close fd)))
    (if flush-only
      (warn "Flush-only/finish-output NYI")
      (progn
        (setf (fblock.fd fblock) nil)
        (let* ((bufvector (fblock.bufvec fblock)))
          (if bufvector
            (%dispose-heap-ivector bufvector)
            (free (fblock.buffer fblock))))
        (setf (fblock.buffer fblock) nil
              (fblock.bufvec fblock) nil)))
    nil))


;(%funtyi fblock char)
;returns nil
; unread char for file referenced by fblock

(defun %funtyi (fblock char)
  (setf (fblock.lastchar fblock) (if (characterp char) char (%code-char char)))
  nil)


;(%feofp fblock)
;returns T if file referenced by fblock is at end of file
;otherwise returns nil

(defun %feofp (fblock)
  (eql (%fsize fblock) (%fpos fblock)))

;(%fpos fblock &optional position)
;returns element position in file.
;if position is supplied, it sets the element position to position and
;returns new position.

; from patch 2 - add truncate arg
(defun %fpos (fblock &optional position truncate)
  (let* ((current8bitbase (fblock.filepos fblock))
         (element-size (fblock.nbits-per-element fblock))
         (logicalpos (+ (%octets-to-elements current8bitbase element-size) (fblock.bufidx fblock))))
    (if (not position)
      (- logicalpos (if (fblock.lastchar fblock) 1 0))
      (let* ((newidx (mod position (fblock.elements-per-buffer fblock)))
             (new8bitbase (%elements-to-octets (- position newidx) element-size))
             eof)
        (setf (fblock.lastchar fblock) nil)
        (setq eof (%fsize fblock))
        (when (> eof (fblock.fileeof fblock))
          (%fsize fblock eof))
        (unless (and (= current8bitbase new8bitbase)
                     (< newidx (fblock.bufcount fblock)))
          (let* ((size (%fsize fblock)))
            (if (> position size)
              (setq position size newidx (mod size (fblock.elements-per-buffer fblock)))))
          (%fbflush fblock)
          (%fbread fblock new8bitbase))
        (setf (fblock.bufidx fblock) newidx)
        (when truncate (setf (fblock.bufcount fblock) newidx))  ; << 6/20/95
        (unless (fblock.bufvec fblock)
          (setf (fblock.element-bit-offset fblock) (* element-size newidx)))
        position))))

;(%fsize fblock &optional new-size)
;returns element size of file referenced by fblock.
;if new-pos is supplied, it sets the file size and returns new-pos.
;if size is set to less than the current file position the file is
;truncated and THE CALLER sets the position to the new size.
;(This is how things have behaved for the last N years, despite
; what the comments might have said...)

(defun %fsize (fblock &optional new-pos)
  (if (null new-pos)
    (let* ((eof (fblock.fileeof fblock))
           (eof1 (if (fblock.dirty fblock)
                   (+ (%octets-to-elements (fblock.filepos fblock) (fblock.nbits-per-element fblock))
                      (max (fblock.bufcount fblock)
                           (fblock.bufidx fblock)))
                   0)))
      (when (fblock.lastchar fblock)
        (decf eof1))
      (max eof eof1))
    (let* ((fd (%fblock.fd fblock))
           (nbits-per-element (fblock.nbits-per-element fblock))
           (offset (- new-pos (%octets-to-elements (fblock.filepos fblock) nbits-per-element))))
      (if (and (>= offset 0) (< offset (fblock.bufsize fblock)))
        (setf (fblock.bufcount fblock) offset))
      (setf (fblock.fileeof fblock) new-pos)
      (fd-ftruncate fd (%elements-to-octets new-pos nbits-per-element))
      new-pos)))


;(%ftyi fblock)
;returns next character from file referenced by fblock.
;if eof, returns nil.

(defun %ftyi (fblock)
  (let* ((ch (fblock.lastchar fblock)))
    (if ch
      (setf (fblock.lastchar fblock) nil)
      (let* ((idx (fblock.bufidx fblock))
             (count (fblock.bufcount fblock))
             (vector (fblock.bufvec fblock)))
        (declare (fixnum idx count))
        (when (< idx count)
          (when vector
            (setq ch (%schar vector idx))
            (setf (fblock.bufidx fblock) (incf idx))
            (if (= idx count)
              (%fbReadAhead0 fblock nil))))))
    ch))
 
  

(defun %fread-byte (fblock)
  (let* ((vector (fblock.bufvec fblock))
         (idx (fblock.bufidx fblock))
         (count (fblock.bufcount fblock)))
    (declare (fixnum idx count))
    (when (< idx count)
      (let* ((byte (if vector (uvref vector idx))))
        (unless byte
          (let* ((bit-offset (fblock.element-bit-offset fblock))
                 (element-size (fblock.nbits-per-element fblock))
                 (unsigned (eql 0 (fblock.minval fblock)))
                 (buffer (fblock.buffer fblock)))
            (declare (fixnum bit-offset element-size))
            (when buffer
              ; here for weird sizes presumably less than 32
              (setf (fblock.element-bit-offset fblock) (the fixnum (+ bit-offset element-size)))
              (setq byte (read-weird-byte bit-offset element-size buffer unsigned)))))
        (setf (fblock.bufidx fblock) (incf idx))
        (if (= idx count)
          (%fbReadAheaD0 fblock nil))
        byte))))


; It's awfully tempting to convert these to LAP...
; Makes you appreciate the 68K bit field instructions.
(defun read-weird-byte (bit-offset element-size buffer unsigned)
  (declare (fixnum bit-offset element-size)
           (type macptr buffer))
  (multiple-value-bind (byte-offset bit-offset-in-byte) (%fixnum-truncate bit-offset 8)
    (declare (fixnum byte-offset bit-offset-in-byte))
    (let* ((initial-bits (- 8 bit-offset-in-byte))
           (initial-mask (1- (the fixnum (%ilsl initial-bits 1))))
           (bits-remaining (- element-size initial-bits))
           (res (the fixnum (logand (the fixnum (%get-unsigned-byte buffer byte-offset)) initial-mask))))
      (declare (fixnum initial-bits initial-mask bits-remaining))
      (if (< bits-remaining 0)
        (%ilsr (the fixnum (- bits-remaining)) res)
        (macrolet ((loop-body (fixnum?)
                     (flet ((left-shift (num bits)
                              (if fixnum?
                                `(%ilsl ,bits ,num)
                                `(ash ,num ,bits))))
                       (let ((type (if fixnum? 'fixnum 'integer)))
                         `(locally (declare ((type ,type) res))
                            (loop
                              (when (eql bits-remaining 0) (return))
                              (incf byte-offset)
                              (when (< bits-remaining 8)
                                (let ((final-shift (- 8 bits-remaining)))
                                  (declare (fixnum final-shift))
                                  (setq res (+ (the ,type ,(left-shift 'res 'bits-remaining))
                                               (the fixnum (%ilsr final-shift (%get-unsigned-byte buffer byte-offset))))))
                                (return))
                              (setq res (+ (the ,type ,(left-shift 'res 8))
                                           (the fixnum (%get-byte buffer byte-offset))))
                              (decf bits-remaining 8))
                            (if unsigned
                              res
                              (if (,(if fixnum? '%ilogbitp 'logbitp) (the fixnum (1- element-size)) res)
                                (the ,type (- res (the ,type ,(left-shift 1 'element-size))))
                                res)))))))
          (if (< element-size (- arch::nbits-in-word arch::fixnum-shift))
            (loop-body t)                 ; guaranteed fixnum result
            (loop-body nil)))))))

; This writes the bits from lsb to msb in byte, the opposite of read-weird-byte
(defun write-weird-byte (bit-offset element-size buffer byte)
  (let ((fixnum-byte (fixnump byte)))
    (multiple-value-bind (byte-offset initial-bits)
                         (%fixnum-truncate (the fixnum (+ bit-offset element-size)) 8)
      (declare (fixnum byte-offset initial-bits))
      (let ((bits-left (- element-size initial-bits)))
        (declare (fixnum bits-left))
        (unless (eql 0 initial-bits)
          (let* ((left-shift (- 8 initial-bits))
                 (mask (1- (the fixnum (%ilsl initial-bits 1))))
                 (read-mask (1- (the fixnum (%ilsl left-shift 1))))
                 (bits (%ilsl left-shift
                              (if fixnum-byte
                                (logand (the fixnum byte) mask)
                                (logand byte mask))))
                 (read-bits (%get-unsigned-byte buffer byte-offset)))
            (declare (fixnum left-shift mask read-mask bits read-bits))
            (when (< bits-left 0)
              (let* ((bits-left (- bits-left))
                     (high-mask (%ilsl (the fixnum (- 8 bits-left))
                                       (the fixnum (1- (%ilsl bits-left 1)))))
                     (read-mask (logior read-mask high-mask))
                     (mask (lognot read-mask)))
                (declare (fixnum shift high-mask read-mask mask))
                (setf (%get-unsigned-byte buffer byte-offset)
                      (logior (the fixnum (logand mask bits))
                              (the fixnum (logand read-bits read-mask)))))
              (return-from write-weird-byte nil))
            (setf (%get-unsigned-byte buffer byte-offset)
                  (logior bits (the fixnum (logand  read-bits read-mask))))
            (setq byte
                  (if fixnum-byte
                    (%ilsr initial-bits byte)
                    (prog1 (ash byte (the fixnum (- initial-bits)))
                      (when (fixnump byte) (setq fixnum-byte t)))))))
        (loop
          (when (eql bits-left 0) (return))
          (decf byte-offset)
          (when (< bits-left 8)
            (let* ((mask (1- (the fixnum (%ilsl bits-left 1))))
                   (read-mask (%ilsl bits-left
                                     (the fixnum
                                       (1- (the fixnum
                                             (%ilsl (the fixnum (- 8 bits-left)) 1))))))
                   (bits (if fixnum-byte
                           (logand (the fixnum byte) mask)
                           (logand byte mask)))
                   (read-bits (logand (the fixnum (%get-unsigned-byte buffer byte-offset)) read-mask)))
              (declare (fixnum mask read-mask bits read-bits))
              (setf (%get-unsigned-byte buffer byte-offset)
                    (the fixnum (+ bits read-bits))))
            (return))
          (let ((bits (if fixnum-byte
                        (prog1 (logand (the fixnum byte) #xff)
                          (setq byte (%ilsr 8 byte)))
                        (prog1 (logand byte #xff)
                          (when (fixnump (setq byte (ash byte -8)))
                            (setq fixnum-byte t))))))
            (setf (%get-unsigned-byte buffer byte-offset) bits))
          (decf bits-left 8))))))



(defun %fbReadAhead0 (fblock writing?)
  (let* ((element-count (fblock.bufsize fblock))
         (nextbase (+ (fblock.filepos fblock) element-count))
         (eof (%elements-to-octets (fblock.fileeof fblock) (fblock.nbits-per-element fblock)))
         (before-eof? (> eof nextbase)))
    (%fbflush fblock)
    (when (or writing? before-eof?)
      (if before-eof?
        (%fbread fblock nextbase)
        (progn (setf (fblock.filepos fblock) nextbase   ; writing past eof.
                     (fblock.bufcount fblock) 0
                     (fblock.bufidx fblock) 0) 
               (unless (fblock.bufvec fblock)
                 (setf (fblock.element-bit-offset fblock) 0)))))))
                   
;(%ftyo fblock char)
;returns nil
;writes char to file referenced by fblock

(defun %ftyo (fblock char)
  (let* ((index (fblock.bufidx fblock))
         (bufsize (fblock.bufsize fblock)))
    (declare (fixnum index bufsize))
    (when (fblock.lastchar fblock)
      (setf (fblock.lastchar fblock) nil)
      (when (< (decf index) 0)
        (%fpos fblock (+ (fblock.filepos fblock) index))
        (setq index (fblock.bufidx fblock))))
    (when (<= bufsize index)
      ; this happens if we just READ the last elt in the buffer and file
      (setf (fblock.bufidx fblock) (setq index 0))
      (incf (fblock.filepos fblock) bufsize)
      (if (<= (fblock.filepos fblock)(fblock.fileeof fblock)) ; <<< 
        (setf (fblock.bufcount fblock) 0)
        (error "Shouldn't in %ftyo. We ought to read the next block first?")))
    (setf (%schar (fblock.bufvec fblock) index) char
          (fblock.bufidx fblock) (incf index)
          (fblock.dirty fblock) t)
    (if (eql index bufsize)
      (%fbreadAHead0 fblock t))
    nil))

#|
; at the end stream-position and file-length should be the same
; This used to fail.
(defun foop ()
  (with-open-file (s "ccl:foo" :direction :io :if-exists :supersede)
    (let ((fblock (slot-value s 'fblock)))
      (dotimes (i 4096)(tyo #\a s))
      (pr-fblock fblock)
      (stream-position s (1- (stream-position s)))
      (pr-fblock fblock)
      (tyi s)
      (format t "~&len ~s" (file-length s))
      (pr-fblock fblock)
      (tyo #\newline s)
      (pr-fblock fblock) ; count was 2048 pos 4096 - wrong
      (format t "~&len ~s" (file-length s))
      (stream-position s))))

(defun pr-fblock (fblock)
  (format t "~&count ~s idx ~s eof ~s pos ~s ~s"
          (fblock.bufcount fblock)
          (fblock.bufidx fblock)
          (fblock.fileeof fblock)
          (fblock.filepos fblock)
          (fblock.dirty fblock)))
|#

(defun %fwrite-byte (fblock byte)
  (let* ((index (fblock.bufidx fblock))
         (bufsize (fblock.elements-per-buffer fblock))
         (min (fblock.minval fblock))
         (max (fblock.maxval fblock)))
    (declare (fixnum index bufsize))
    (unless (and (>= byte min) (<= byte max))
      (report-bad-arg byte `(integer ,min ,max)))
    (when (<= bufsize index)
      ; this happens if we just READ the last elt in the buffer and file
      (setf (fblock.bufidx fblock) (setq index 0))
      (incf (fblock.filepos fblock) bufsize)
      (if (<= (fblock.filepos fblock)(fblock.fileeof fblock)) ; <<< 
        (setf (fblock.bufcount fblock) 0)
        (error "Shouldn't in %fwrite-byte. We ought to read the next block first?")))    
    (let* ((vector (fblock.bufvec fblock)))
      (if vector 
        (setf (uvref vector index) byte)
        (let* ((bit-offset (fblock.element-bit-offset fblock))
               (element-size (fblock.nbits-per-element fblock))
               (buffer (fblock.buffer fblock)))
          (declare (fixnum bit-offset element-size))
          (when buffer
            (setf (fblock.element-bit-offset fblock) (the fixnum (+ bit-offset element-size)))
            (write-weird-byte bit-offset element-size buffer byte)))))
    (setf (fblock.bufidx fblock) (incf index)
          (fblock.dirty fblock) t)
    (if (eql index bufsize)
      (%fbreadAHead0 fblock t))))

(defun %fwrite-from-vector (fblock vector start length)
    (unless (= 8 (the fixnum (fblock.nbits-per-element fblock)))
      (error "Can't write vector to file ~s" (fblock.stream fblock)))
    (let* ((written 0)
           (bufsize (fblock.elements-per-buffer fblock))
           (buffer (fblock.buffer fblock)))
      (declare (fixnum written bufsize))
      (do* ((pos start (+ pos written))
            (left length (- left written)))
           ((= left 0) length)
        (declare (fixnum pos left))
        (setf (fblock.dirty fblock) t)
         (let* ((index (fblock.bufidx fblock))
                (avail (- bufsize index)))
          (declare (fixnum avail index))
          (cond 
           ((= (setq written avail) 0)
            (%fbreadahead0 fblock t))
           (t 
            (if (> written left)
              (setq written left))
            (%copy-ivector-to-ptr vector pos buffer index written)            
            (if (= (setf (fblock.bufidx fblock) (+ index written)) bufsize)
              (%fbreadahead0 fblock t))))))))

(defun %fread-to-vector (fblock vector start length)
  (unless (= 8 (the fixnum (fblock.nbits-per-element fblock)))
    (error "Can't read vector from file ~s" (fblock.stream fblock)))
  (if (= (fblock.bufcount fblock) (fblock.bufidx fblock))
    (%fbreadahead0 fblock nil))
  (let* ((nread 0)
         (bufsize (fblock.bufcount fblock))
         (buffer (fblock.buffer fblock)))
    (declare (fixnum nread bufsize))
    (do* ((pos start (+ pos nread))
          (left length (- left nread)))
         ((= left 0) (values vector length))
      (declare (fixnum pos left))
      (let* ((index (fblock.bufidx fblock))
             (avail (- bufsize index)))
        (declare (fixnum avail index))
        (if (zerop (setq nread avail))
          (return (values vector (- length left))))
        (if (> nread left)
          (setq nread left))
        (%copy-ptr-to-ivector buffer index vector pos nread)        
        (when (= (setf (fblock.bufidx fblock) (+ index nread)) bufsize)
          (%fbreadahead0 fblock nil)
          (setq bufsize (fblock.bufcount fblock)))))))

; Do read of a block from BASE address
; Assumes all I/O is done on fd
; Since all of it happens synchronously, that's a pretty safe assumption.
(defun %fbRead (fblock base)
  (let* ((element-size (fblock.nbits-per-element fblock))
         (fd (%fblock.fd fblock)))
    (declare (type fixnum fd))
    
    (setf (fblock.filepos fblock) base)         ; 8-bit
    (fd-lseek fd base $seek-set)
    (let* ((nread (fd-read fd (fblock.buffer fblock) (fblock.bufsize fblock))))
      (when (< nread 0)
        (signal-file-error nread))
      (setf (fblock.bufidx fblock) 0)
      (unless (fblock.bufvec fblock)
        (setf (fblock.element-bit-offset fblock) 0))
      (setf (fblock.bufcount fblock) (%octets-to-elements nread element-size)))))

; Flush buffer if it's dirty.
(defun %fbFlush (fblock)
  (when (fblock.dirty fblock)
    (setf (fblock.dirty fblock) nil)
    (let* ((element-size (fblock.nbits-per-element fblock))
           (fd (%fblock.fd fblock))
           (8-bit-filepos (fblock.filepos fblock))
           (filepos (%octets-to-elements 8-bit-filepos (fblock.nbits-per-element fblock)))
           (elements-written (fblock.bufidx fblock))
           (elements-read (fblock.bufcount fblock))
           (new-eof (+ filepos (max elements-written elements-read))))
      (declare (fixnum fd))
      (when (> new-eof (fblock.fileeof fblock))
        (setf (fblock.fileeof fblock) new-eof))
      (fd-lseek fd 8-bit-filepos $seek-set)
      (let* ((count (%elements-to-octets (max elements-written elements-read) element-size))
             (nwritten (fd-write fd (fblock.buffer fblock) count)))
        (unless (= count nwritten)
          (errchk nwritten))))))

#| Test routines

(defun fopen (file &optional (perm $fsRdWrPerm))
  (%fopen file 0 perm))

(defun pf (fblock &optional max &aux char (count 0))
  (while (setq char (%ftyi fblock))
    (tyo char)
    (incf count)
    (if (and max (>= count max))
      (return-from pf count)))
  count)

(defun fcopy (from to &aux char (count 0))
  (while (setq char (%ftyi from))
    (%ftyo to char)
    (incf count))
  count)

(defun freverse (f)
  (let* ((size (%fsize f))
         (p 0)
         (p2 (1- size))
         c c2)
    (while (> p2 p)
      (%fpos f p)
      (setq c (%ftyi f))
      (%fpos f p2)
      (setq c2 (%ftyi f))
      (%fpos f p2)
      (%ftyo f c)
      (%fpos f p)
      (%ftyo f c2)
      (incf p)
      (decf p2))
    (%fpos f 0)
    (%fsize f)))

; I used different names during development
(defun install-new-sysio ()
  (setf (symbol-function '%fopen) #'%nfopen)
  (setf (symbol-function '%fclose) #'%nfclose)
  (setf (symbol-function '%feofp) #'%nfeofp)
  (setf (symbol-function '%ftyi) #'%nftyi)
  (setf (symbol-function '%ftyo) #'%nftyo)
  (setf (symbol-function '%fpos) #'%nfpos)
  (setf (symbol-function '%fsize) #'%nfsize))

(defun c (&optional (file "l1;new-l1-sysio"))
  (time
   (compile-file file
                 :output-file (concatenate 'string "l1f;" (pathname-name file)))))

|#
