;;;-*-Mode: LISP; Package: CCL -*-

; Copyright 1995 Digitool, Inc.

(in-package :ccl)

; Modification History
;
; 2/15/95 slh  split out from compile-ccl.lisp

(require 'compile-ccl)

(defun relative-pathname (path to)
  (when (and (setq path (probe-file path))
             (setq to (probe-file to)))
    (setq path (namestring path)
          to (namestring to))
    (when (eql 0 (search to path))
      (subseq path (length to) (length path)))))

(defun same-file-write-date-and-length-p (file1 file2 &optional time-delta)
  (ignore-errors
   (multiple-value-bind (cd1 md1 dl1 rl1) (file-dates-and-lengths file1)
     (multiple-value-bind (cd2 md2 dl2 rl2) (file-dates-and-lengths file2)
       (flet ((same-time (time1 time2 time-delta)
                (or (eql time1 time2)
                    (and time-delta
                         (let ((diff (- time1 time2)))
                           (if (listp time-delta)
                             (eql 0 (mod diff (car time-delta)))
                             (or (eql diff time-delta)
                                 (eql diff (- time-delta)))))))))
         (and (same-time cd1 cd2 time-delta)
              (same-time md1 md2 time-delta)
              (eql dl1 dl2)
              (eql rl1 rl2)))))))

; Returns 4 values:
; 1) file creation date as Mac (not universal) time
; 2) file modification date as Mac time
; 3) data fork length
; 4) resource fork length
(defun file-dates-and-lengths (path)
  (%stack-iopb (pb np)
    (%path-to-iopb path pb :errchk)
    (values
     (%get-long pb $ioFlCrDat)
     (%get-long pb $ioFlMdDat)
     (%get-long pb $ioFlLgLen)
     (%get-long pb $ioFlRLgLen))))

(defvar *ccl-system-source-fasls* nil)
(defvar *ccl-system-snapshot* nil)
(def-ccl-pointers *ccl-system-snapshot* ()
  (setq *ccl-system-snapshot* nil))

(defun get-ccl-system-source-fasls-cell (source &optional (consp t))
  (setq source (mac-namestring source))
  (let ((cell (assoc source *ccl-system-source-fasls* :test 'string-equal)))
    (or cell
        (and consp (car (push (cons source nil) *ccl-system-source-fasls*))))))

(defun delete-ccl-system-source-fasls (source)
  (let ((cell (get-ccl-system-source-fasls-cell source nil)))
    (dolist (fasl (cdr cell))
      (when (and fasl (probe-file fasl))
        (format t "~&Deleting ~s" (back-translate-pathname fasl))
        (delete-file fasl)))))

(defun make-ccl-system-source-fasls (&optional build-directory force?)
  (unless (and (not force?)
               *ccl-system-source-fasls*
               (eq *ccl-system-snapshot* *ccl-system*))
    (setq *ccl-system-source-fasls* nil)
    (setq *ccl-system-snapshot* *ccl-system*)
    (dolist (spec *ccl-system-snapshot*)
      (let ((fasl (second spec)))
        (dolist (source (third spec))
          (push fasl (cdr (get-ccl-system-source-fasls-cell source))))))
    (when build-directory
      (let ((obj-dir (merge-pathnames ":lispobj:.o" build-directory)))
        (dolist (file (directory (merge-pathnames ":asms:*.a" build-directory)))
          (push (merge-pathnames obj-dir file)
                (cdr (get-ccl-system-source-fasls-cell file)))))))
  *ccl-system-source-fasls*)

(defun prompt-for-character (chars format-string &rest arguments &aux response)
  (declare (dynamic-extent arguments))
  (loop
    (when format-string
      (fresh-line *query-io*)
      (apply 'format *query-io* format-string arguments))
    (setq response (read-char *query-io*))
    (if (if (functionp chars)
          (apply chars response)
          (find response chars :test #'char-equal))
      (return response))
    (ed-beep)))

(defun merge-ccl-update (update-directory
                         &key
                         (delete-fasls t)
                         (ccl-directory "ccl:")
                         (build-directory (namestring
                                           (or (probe-file "ccl:build;")
                                               (probe-file (merge-pathnames 
                                                            ":build:" 
                                                            (drive-name (drive-number "ccl:"))))
                                               (error "Can't find BUILD directory."))))
                         (rename-vice-copy-files nil)
                         (copy-identical-files t)
                         (yes-mode nil)
                         (start-file nil)
                         (time-delta nil))
  (require :compare)                    ; define compare-files-to-buffer
  (setq ccl-directory (namestring (translate-logical-pathname ccl-directory)))
  (when delete-fasls (make-ccl-system-source-fasls build-directory))
  (when (eql #\: (char (setq update-directory (namestring update-directory)) 0))
    (setq update-directory 
          (concatenate 'string "buttle:CCL Development" update-directory)))
  (unless (eql #\: (char update-directory (1- (length update-directory))))
    (setq update-directory (concatenate 'string update-directory ":")))
  (let ((files (directory (merge-pathnames ":**:*.*" update-directory)))
        file out-file maybe-copy? copy? compare?)
    (when start-file
      (setq files (member (pathname start-file) files
                          :test #'(lambda (start-file file)
                                    (and (equalp (pathname-name start-file)
                                                 (pathname-name file))
                                         (equalp (pathname-type start-file)
                                                 (pathname-type file)))))))
    (loop
      (unless (setq file (pop files)) (return))
      (setq file (namestring file))
      (labels ((out-file (file)
                 (let* ((relative-path (relative-pathname file update-directory))
                        (build-p (eql 0 (search "build:" relative-path :key #'char-downcase))))
                   (concatenate 'string
                                (if build-p build-directory ccl-directory)
                                (if build-p
                                  (subseq relative-path 6 (length relative-path))
                                  relative-path))))
              (files-identical? (file out-file)
                 (when (same-file-write-date-and-length-p file out-file time-delta)
                   (format t "~&Files ~s and~%      ~s appear to be identical." 
                           (bt file) (bt out-file))
                   t))
              (bt (pathname) (back-translate-pathname pathname)))
        (setq out-file (out-file file)
              maybe-copy? t
              copy? nil
              compare? nil)
        (unless (files-identical? file out-file)
          (flet ((delete-fasl (source)
                   (when delete-fasls
                     (delete-ccl-system-source-fasls source))))
            (when (and (eq :text (mac-file-type file))
                       (probe-file out-file))
              (let ((ans (if yes-mode
                           #\Y
                           (prompt-for-character
                            "YNCDSK"
                            "Compare ~s to~%        ~s?~%        (Yes, No, Copy, copy Directory, Skip file, sKip directory) "
                            file out-file))))
                (case (setq ans (char-upcase ans))
                  (#\Y (setq compare? t))
                  (#\N)
                  (#\C (setq copy? t))
                  (#\D (setq copy? #\D))
                  (#\S (setq maybe-copy? nil))
                  (#\K (setq copy? #\K))))
              (when compare?
                (when yes-mode
                  (format t "~&Comparing ~s to~%          ~s~%" file out-file))
                (restart-case
                  (multiple-value-bind (diffs window)
                                       (funcall 'compare-files-to-buffer file out-file)
                    (declare (ignore diffs))
                    (if window
                      (let ((date (file-write-date out-file)))
                        (while (wptr window)
                          (when *eval-queue*
                            (with-simple-restart (abort "Abort eval inside of merge")
                              (catch-cancel
                                (eval (pop *eval-queue*)))))
                          (event-dispatch))
                        (unless (eql date (file-write-date out-file))
                          (setq maybe-copy? nil)
                          (delete-fasl out-file)))
                      (progn
                        (format t "~&Files ~s and ~%      ~s are identical" (bt file) (bt out-file))
                        (if copy-identical-files
                          (let ((r1 (nth-value 3 (file-dates-and-lengths file)))
                                (r2 (nth-value 3 (file-dates-and-lengths out-file))))
                            (if (eql r1 r2)
                              (progn
                                (format t "~&Updating dates of ~s~%" out-file)
                                (set-file-create-date out-file (file-create-date file))
                                (set-file-write-date out-file (file-write-date file))
                                (setq maybe-copy? nil))
                              (setq copy? t)))
                          (setq maybe-copy? nil)))))
                  (skip-compare ()
                                :report (lambda (s) (format s "Skip comparing files"))
                                (setq maybe-copy? t)))))
            (when (or copy? maybe-copy?)
              (unless copy?
                (setq copy? (char-upcase
                             (prompt-for-character
                              "YNDK"
                              "Copy ~s to~%     ~s?~%     (Yes, No, copy Directory, sKip directory) "
                              file out-file)))
                (if (eql #\N copy?) (setq copy? nil)))
              (when copy?
                (flet ((copy (file out-file)
                         (unless (files-identical? file out-file)
                           (create-directory (mac-directory-namestring out-file)
                                             :if-exists nil)
                           (if (and rename-vice-copy-files
                                    (eql (drive-number file) (drive-number out-file)))
                             (progn
                               (format t "~&Renaming ~s to~%         ~s..." 
                                       (bt file) (bt out-file))
                               (rename-file file out-file :if-exists :supersede)
                               (format t "Done.~%"))
                             (progn
                               (format t "~&Copying ~s to~%        ~s..." (bt file) (bt out-file))
                               (copy-file file out-file :if-exists :supersede)
                               (format t "Done.~%")))
                           (delete-fasl out-file))))
                  (cond ((eql #\D copy?)
                         (progn
                           (copy file out-file)
                           (dolist (file (directory
                                          (merge-pathnames 
                                           ":**:*.*" (mac-directory-namestring file))))
                             (when (member file files :test #'equal)
                               (setq files (delete file files :test #'equal))
                               (copy file (out-file file))))))
                        ((eql #\K copy?)
                         (dolist (file (directory
                                        (merge-pathnames 
                                         ":**:*.*" (mac-directory-namestring file))))
                           (setq files (delete file files :test #'equal))))
                        (t (copy file out-file))))))))))))

; End of merge-ccl-update.lisp
